@@ -1,4 +1,4 @@
-/**
+/*
  * Licensed to the Apache Software Foundation (ASF) under one
  * or more contributor license agreements.  See the NOTICE file
  * distributed with this work for additional information
@@ -22,11 +22,12 @@ import static org.testng.Assert.assertEquals;
 import static org.testng.Assert.assertFalse;
 import static org.testng.Assert.assertTrue;
 import static org.testng.Assert.fail;
-import com.beust.jcommander.internal.Maps;
 import com.google.common.collect.Sets;
 import java.net.URL;
 import java.time.Duration;
 import java.util.ArrayList;
+import java.util.Collection;
+import java.util.HashMap;
 import java.util.List;
 import java.util.Optional;
 import java.util.UUID;
@@ -35,9 +36,11 @@ import java.util.concurrent.CyclicBarrier;
 import java.util.concurrent.TimeUnit;
 import java.util.concurrent.atomic.AtomicBoolean;
 import lombok.Cleanup;
+import org.apache.bookkeeper.mledger.Position;
 import org.apache.bookkeeper.mledger.impl.ManagedLedgerImpl;
 import org.apache.pulsar.broker.PulsarService;
 import org.apache.pulsar.broker.ServiceConfiguration;
+import org.apache.pulsar.broker.auth.MockedPulsarServiceBaseTest;
 import org.apache.pulsar.broker.service.persistent.PersistentTopic;
 import org.apache.pulsar.client.admin.GetStatsOptions;
 import org.apache.pulsar.client.admin.PulsarAdmin;
@@ -55,6 +58,7 @@ import org.apache.pulsar.common.policies.data.ClusterData;
 import org.apache.pulsar.common.policies.data.PersistentTopicInternalStats;
 import org.apache.pulsar.common.policies.data.TenantInfoImpl;
 import org.apache.pulsar.common.policies.data.TopicStats;
+import org.apache.pulsar.common.policies.data.TopicType;
 import org.apache.pulsar.common.policies.data.impl.BacklogQuotaImpl;
 import org.apache.pulsar.zookeeper.LocalBookkeeperEnsemble;
 import org.awaitility.Awaitility;
@@ -80,6 +84,22 @@ public class BacklogQuotaManagerTest {
     private static final int TIME_TO_CHECK_BACKLOG_QUOTA = 2;
     private static final int MAX_ENTRIES_PER_LEDGER = 5;
 
+    /**
+     * see {@link MockedPulsarServiceBaseTest#deleteNamespaceWithRetry(String, boolean, PulsarAdmin, Collection)}
+     */
+    protected void deleteNamespaceWithRetry(String ns, boolean force)
+            throws Exception {
+        MockedPulsarServiceBaseTest.deleteNamespaceWithRetry(ns, force, admin, pulsar);
+    }
+
+    /**
+     * see {@link MockedPulsarServiceBaseTest#deleteNamespaceWithRetry(String, boolean, PulsarAdmin, Collection)}
+     */
+    protected void deleteNamespaceWithRetry(String ns, boolean force, PulsarAdmin admin)
+            throws Exception {
+        MockedPulsarServiceBaseTest.deleteNamespaceWithRetry(ns, force, admin, pulsar);
+    }
+
     @DataProvider(name = "backlogQuotaSizeGB")
     public Object[][] backlogQuotaSizeGB() {
         return new Object[][] { { true }, { false } };
@@ -94,18 +114,21 @@ public class BacklogQuotaManagerTest {
 
             // start pulsar service
             config = new ServiceConfiguration();
-            config.setZookeeperServers("127.0.0.1" + ":" + bkEnsemble.getZookeeperPort());
+            config.setMetadataStoreUrl("zk:127.0.0.1:" + bkEnsemble.getZookeeperPort());
             config.setAdvertisedAddress("localhost");
             config.setWebServicePort(Optional.of(0));
             config.setClusterName("usc");
             config.setBrokerShutdownTimeoutMs(0L);
+            config.setLoadBalancerOverrideBrokerNicSpeedGbps(Optional.of(1.0d));
             config.setBrokerServicePort(Optional.of(0));
             config.setAuthorizationEnabled(false);
             config.setAuthenticationEnabled(false);
             config.setBacklogQuotaCheckIntervalInSeconds(TIME_TO_CHECK_BACKLOG_QUOTA);
             config.setManagedLedgerMaxEntriesPerLedger(MAX_ENTRIES_PER_LEDGER);
             config.setManagedLedgerMinLedgerRolloverTimeMinutes(0);
-            config.setAllowAutoTopicCreationType("non-partitioned");
+            config.setAllowAutoTopicCreationType(TopicType.NON_PARTITIONED);
+            config.setSystemTopicEnabled(true);
+            config.setTopicLevelPoliciesEnabled(true);
             config.setForceDeleteNamespaceAllowed(true);
 
             pulsar = new PulsarService(config);
@@ -156,10 +179,10 @@ public class BacklogQuotaManagerTest {
     }
 
     @AfterMethod(alwaysRun = true)
-    void clearNamespaces() throws PulsarAdminException {
-        admin.namespaces().deleteNamespace("prop/ns-quota", true);
-        admin.namespaces().deleteNamespace("prop/quotahold", true);
-        admin.namespaces().deleteNamespace("prop/quotaholdasync", true);
+    void clearNamespaces() throws Exception {
+        deleteNamespaceWithRetry("prop/ns-quota", true);
+        deleteNamespaceWithRetry("prop/quotahold", true);
+        deleteNamespaceWithRetry("prop/quotaholdasync", true);
     }
 
     private void rolloverStats() {
@@ -172,7 +195,7 @@ public class BacklogQuotaManagerTest {
     @Test
     public void testBacklogQuotaWithReader() throws Exception {
         assertEquals(admin.namespaces().getBacklogQuotaMap("prop/ns-quota"),
-                Maps.newHashMap());
+                new HashMap<>());
         admin.namespaces().setBacklogQuota("prop/ns-quota",
                 BacklogQuota.builder()
                         .limitSize(10 * 1024)
@@ -180,7 +203,7 @@ public class BacklogQuotaManagerTest {
                         .retentionPolicy(BacklogQuota.RetentionPolicy.producer_exception)
                         .build());
         try (PulsarClient client = PulsarClient.builder().serviceUrl(adminUrl.toString()).statsInterval(0, TimeUnit.SECONDS).build();) {
-            final String topic1 = "persistent://prop/ns-quota/topic1";
+            final String topic1 = "persistent://prop/ns-quota/topic1" + UUID.randomUUID();
             final int numMsgs = 20;
 
             Reader<byte[]> reader = client.newReader().topic(topic1).receiverQueueSize(1).startMessageId(MessageId.latest).create();
@@ -257,7 +280,7 @@ public class BacklogQuotaManagerTest {
     @Test
     public void testTriggerBacklogQuotaSizeWithReader() throws Exception {
         assertEquals(admin.namespaces().getBacklogQuotaMap("prop/ns-quota"),
-                Maps.newHashMap());
+                new HashMap<>());
         admin.namespaces().setBacklogQuota("prop/ns-quota",
                 BacklogQuota.builder()
                         .limitSize(10 * 1024)
@@ -329,7 +352,7 @@ public class BacklogQuotaManagerTest {
     @Test
     public void testTriggerBacklogTimeQuotaWithReader() throws Exception {
         assertEquals(admin.namespaces().getBacklogQuotaMap("prop/ns-quota"),
-                Maps.newHashMap());
+                new HashMap<>());
         admin.namespaces().setBacklogQuota("prop/ns-quota",
                 BacklogQuota.builder()
                         .limitSize(10 * 1024)
@@ -396,7 +419,7 @@ public class BacklogQuotaManagerTest {
     @Test
     public void testConsumerBacklogEvictionSizeQuota() throws Exception {
         assertEquals(admin.namespaces().getBacklogQuotaMap("prop/ns-quota"),
-                Maps.newHashMap());
+                new HashMap<>());
         admin.namespaces().setBacklogQuota("prop/ns-quota",
                 BacklogQuota.builder()
                         .limitSize(10 * 1024)
@@ -406,7 +429,7 @@ public class BacklogQuotaManagerTest {
         PulsarClient client = PulsarClient.builder().serviceUrl(adminUrl.toString()).statsInterval(0, TimeUnit.SECONDS)
                 .build();
 
-        final String topic1 = "persistent://prop/ns-quota/topic2";
+        final String topic1 = "persistent://prop/ns-quota/topic2" + UUID.randomUUID();
         final String subName1 = "c1";
         final String subName2 = "c2";
         final int numMsgs = 20;
@@ -431,17 +454,18 @@ public class BacklogQuotaManagerTest {
     @Test
     public void testConsumerBacklogEvictionTimeQuotaPrecise() throws Exception {
         assertEquals(admin.namespaces().getBacklogQuotaMap("prop/ns-quota"),
-                Maps.newHashMap());
+                new HashMap<>());
         admin.namespaces().setBacklogQuota("prop/ns-quota",
                 BacklogQuota.builder()
                         .limitTime(TIME_TO_CHECK_BACKLOG_QUOTA)
                         .retentionPolicy(BacklogQuota.RetentionPolicy.consumer_backlog_eviction)
                         .build(), BacklogQuota.BacklogQuotaType.message_age);
         config.setPreciseTimeBasedBacklogQuotaCheck(true);
+        @Cleanup
         PulsarClient client = PulsarClient.builder().serviceUrl(adminUrl.toString()).statsInterval(0, TimeUnit.SECONDS)
                 .build();
 
-        final String topic1 = "persistent://prop/ns-quota/topic3";
+        final String topic1 = "persistent://prop/ns-quota/topic3" + UUID.randomUUID();
         final String subName1 = "c1";
         final String subName2 = "c2";
         final int numMsgs = 9;
@@ -467,22 +491,22 @@ public class BacklogQuotaManagerTest {
         // All messages for both subscription should be cleaned up from backlog by backlog monitor task.
         assertEquals(stats.getSubscriptions().get(subName1).getMsgBacklog(), 0);
         assertEquals(stats.getSubscriptions().get(subName2).getMsgBacklog(), 0);
-        client.close();
     }
 
-    @Test
+    @Test(timeOut = 60000)
     public void testConsumerBacklogEvictionTimeQuota() throws Exception {
         assertEquals(admin.namespaces().getBacklogQuotaMap("prop/ns-quota"),
-                Maps.newHashMap());
+                new HashMap<>());
         admin.namespaces().setBacklogQuota("prop/ns-quota",
                 BacklogQuota.builder()
                         .limitTime(TIME_TO_CHECK_BACKLOG_QUOTA)
                         .retentionPolicy(BacklogQuota.RetentionPolicy.consumer_backlog_eviction)
                         .build(), BacklogQuota.BacklogQuotaType.message_age);
+        @Cleanup
         PulsarClient client = PulsarClient.builder().serviceUrl(adminUrl.toString()).statsInterval(0, TimeUnit.SECONDS)
                 .build();
 
-        final String topic1 = "persistent://prop/ns-quota/topic3";
+        final String topic1 = "persistent://prop/ns-quota/topic3" + UUID.randomUUID();
         final String subName1 = "c1";
         final String subName2 = "c2";
         final int numMsgs = 14;
@@ -501,33 +525,93 @@ public class BacklogQuotaManagerTest {
         assertEquals(stats.getSubscriptions().get(subName1).getMsgBacklog(), 14);
         assertEquals(stats.getSubscriptions().get(subName2).getMsgBacklog(), 14);
 
+        PersistentTopic topic1Reference = (PersistentTopic) pulsar.getBrokerService().getTopicReference(topic1).get();
+        ManagedLedgerImpl ml = (ManagedLedgerImpl) topic1Reference.getManagedLedger();
+        Position slowConsumerReadPos = ml.getSlowestConsumer().getReadPosition();
+
         Thread.sleep((TIME_TO_CHECK_BACKLOG_QUOTA * 2) * 1000);
         rolloverStats();
 
-        stats = getTopicStats(topic1);
-        PersistentTopic topic1Reference = (PersistentTopic) pulsar.getBrokerService().getTopicReference(topic1).get();
-        ManagedLedgerImpl ml = (ManagedLedgerImpl) topic1Reference.getManagedLedger();
+        TopicStats stats2 = getTopicStats(topic1);
         // Messages on first 2 ledgers should be expired, backlog is number of
         // message in current ledger.
-        assertEquals(stats.getSubscriptions().get(subName1).getMsgBacklog(), ml.getCurrentLedgerEntries());
-        assertEquals(stats.getSubscriptions().get(subName2).getMsgBacklog(), ml.getCurrentLedgerEntries());
-        client.close();
+        Awaitility.await().untilAsserted(() -> {
+            assertEquals(stats2.getSubscriptions().get(subName1).getMsgBacklog(), ml.getCurrentLedgerEntries());
+            assertEquals(stats2.getSubscriptions().get(subName2).getMsgBacklog(), ml.getCurrentLedgerEntries());
+        });
+
+        assertEquals(ml.getSlowestConsumer().getReadPosition(), slowConsumerReadPos);
     }
 
+    @Test(timeOut = 60000)
+    public void testConsumerBacklogEvictionTimeQuotaWithPartEviction() throws Exception {
+        assertEquals(admin.namespaces().getBacklogQuotaMap("prop/ns-quota"),
+                new HashMap<>());
+        admin.namespaces().setBacklogQuota("prop/ns-quota",
+                BacklogQuota.builder()
+                        .limitTime(5) // set limit time as 5 seconds
+                        .retentionPolicy(BacklogQuota.RetentionPolicy.consumer_backlog_eviction)
+                        .build(), BacklogQuota.BacklogQuotaType.message_age);
+        @Cleanup
+        PulsarClient client = PulsarClient.builder().serviceUrl(adminUrl.toString()).statsInterval(0, TimeUnit.SECONDS)
+                .build();
+
+        final String topic1 = "persistent://prop/ns-quota/topic3" + UUID.randomUUID();
+        final String subName1 = "c1";
+        final String subName2 = "c2";
+        int numMsgs = 5;
+
+        Consumer<byte[]> consumer1 = client.newConsumer().topic(topic1).subscriptionName(subName1).subscribe();
+        Consumer<byte[]> consumer2 = client.newConsumer().topic(topic1).subscriptionName(subName2).subscribe();
+        org.apache.pulsar.client.api.Producer<byte[]> producer = createProducer(client, topic1);
+        byte[] content = new byte[1024];
+        for (int i = 0; i < numMsgs; i++) {
+            producer.send(content);
+            consumer1.receive();
+            consumer2.receive();
+        }
+
+        TopicStats stats = getTopicStats(topic1);
+        assertEquals(stats.getSubscriptions().get(subName1).getMsgBacklog(), 5);
+        assertEquals(stats.getSubscriptions().get(subName2).getMsgBacklog(), 5);
+
+        // Sleep 5000 mills for first 5 messages.
+        Thread.sleep(5000L);
+        numMsgs = 9;
+        for (int i = 0; i < numMsgs; i++) {
+            producer.send(content);
+            consumer1.receive();
+            consumer2.receive();
+        }
+
+        // The first 5 messages are expired after sleeping 2000 more mills.
+        Thread.sleep(2000L);
+        rolloverStats();
+
+        TopicStats stats2 = getTopicStats(topic1);
+        // The first 5 messages should be expired due to limit time is 5 seconds, and the last 9 message should not.
+        Awaitility.await().untilAsserted(() -> {
+            assertEquals(stats2.getSubscriptions().get(subName1).getMsgBacklog(), 9);
+            assertEquals(stats2.getSubscriptions().get(subName2).getMsgBacklog(), 9);
+        });
+    }
+
+
     @Test
     public void testConsumerBacklogEvictionTimeQuotaWithEmptyLedger() throws Exception {
         assertEquals(admin.namespaces().getBacklogQuotaMap("prop/ns-quota"),
-                Maps.newHashMap());
+                new HashMap<>());
         admin.namespaces().setBacklogQuota("prop/ns-quota",
                 BacklogQuota.builder()
                         .limitTime(TIME_TO_CHECK_BACKLOG_QUOTA)
                         .retentionPolicy(BacklogQuota.RetentionPolicy.consumer_backlog_eviction)
                         .build(), BacklogQuota.BacklogQuotaType.message_age);
         config.setPreciseTimeBasedBacklogQuotaCheck(true);
+        @Cleanup
         PulsarClient client = PulsarClient.builder().serviceUrl(adminUrl.toString()).statsInterval(0, TimeUnit.SECONDS)
                 .build();
 
-        final String topic = "persistent://prop/ns-quota/topic4";
+        final String topic = "persistent://prop/ns-quota/topic4" + UUID.randomUUID();
         final String subName = "c1";
 
         Consumer<byte[]> consumer = client.newConsumer().topic(topic).subscriptionName(subName).subscribe();
@@ -557,23 +641,22 @@ public class BacklogQuotaManagerTest {
             TopicStats latestStats = getTopicStats(topic);
             assertEquals(latestStats.getSubscriptions().get(subName).getMsgBacklog(), 0);
         });
-
-        client.close();
     }
 
     @Test
     public void testConsumerBacklogEvictionWithAckSizeQuota() throws Exception {
         assertEquals(admin.namespaces().getBacklogQuotaMap("prop/ns-quota"),
-                Maps.newHashMap());
+                new HashMap<>());
         admin.namespaces().setBacklogQuota("prop/ns-quota",
                 BacklogQuota.builder()
                         .limitSize(10 * 1024)
                         .limitTime(TIME_TO_CHECK_BACKLOG_QUOTA)
                         .retentionPolicy(BacklogQuota.RetentionPolicy.consumer_backlog_eviction)
                         .build());
+        @Cleanup
         PulsarClient client = PulsarClient.builder().serviceUrl(adminUrl.toString()).build();
 
-        final String topic1 = "persistent://prop/ns-quota/topic11";
+        final String topic1 = "persistent://prop/ns-quota/topic11" + UUID.randomUUID();
         final String subName1 = "c11";
         final String subName2 = "c21";
         final int numMsgs = 20;
@@ -599,16 +682,17 @@ public class BacklogQuotaManagerTest {
     @Test
     public void testConsumerBacklogEvictionWithAckTimeQuotaPrecise() throws Exception {
         assertEquals(admin.namespaces().getBacklogQuotaMap("prop/ns-quota"),
-                Maps.newHashMap());
+                new HashMap<>());
         admin.namespaces().setBacklogQuota("prop/ns-quota",
                 BacklogQuota.builder()
                         .limitTime(TIME_TO_CHECK_BACKLOG_QUOTA)
                         .retentionPolicy(BacklogQuota.RetentionPolicy.consumer_backlog_eviction)
                         .build(), BacklogQuota.BacklogQuotaType.message_age);
         config.setPreciseTimeBasedBacklogQuotaCheck(true);
+        @Cleanup
         PulsarClient client = PulsarClient.builder().serviceUrl(adminUrl.toString()).build();
 
-        final String topic1 = "persistent://prop/ns-quota/topic12";
+        final String topic1 = "persistent://prop/ns-quota/topic12" + UUID.randomUUID();
         final String subName1 = "c11";
         final String subName2 = "c21";
         final int numMsgs = 9;
@@ -647,7 +731,6 @@ public class BacklogQuotaManagerTest {
         stats = getTopicStats(topic1);
         // sub2 has empty backlog because it's backlog get cleaned up by backlog quota monitor task
         assertEquals(stats.getSubscriptions().get(subName2).getMsgBacklog(), 0);
-        client.close();
     }
 
     private Producer<byte[]> createProducer(PulsarClient client, String topic)
@@ -662,11 +745,11 @@ public class BacklogQuotaManagerTest {
     @Test
     public void testConsumerBacklogEvictionWithAckTimeQuota() throws Exception {
         assertEquals(admin.namespaces().getBacklogQuotaMap("prop/ns-quota"),
-                Maps.newHashMap());
+                new HashMap<>());
         @Cleanup
         PulsarClient client = PulsarClient.builder().serviceUrl(adminUrl.toString()).build();
 
-        final String topic1 = "persistent://prop/ns-quota/topic12";
+        final String topic1 = "persistent://prop/ns-quota/topic12" + UUID.randomUUID();
         final String subName1 = "c11";
         final String subName2 = "c21";
         final int numMsgs = 14;
@@ -732,14 +815,14 @@ public class BacklogQuotaManagerTest {
     @Test
     public void testConcurrentAckAndEviction() throws Exception {
         assertEquals(admin.namespaces().getBacklogQuotaMap("prop/ns-quota"),
-                Maps.newHashMap());
+                new HashMap<>());
         admin.namespaces().setBacklogQuota("prop/ns-quota",
                 BacklogQuota.builder()
                         .limitSize(10 * 1024)
                         .retentionPolicy(BacklogQuota.RetentionPolicy.consumer_backlog_eviction)
                         .build());
 
-        final String topic1 = "persistent://prop/ns-quota/topic12";
+        final String topic1 = "persistent://prop/ns-quota/topic12" + UUID.randomUUID();
         final String subName1 = "c12";
         final String subName2 = "c22";
         final int numMsgs = 20;
@@ -756,40 +839,36 @@ public class BacklogQuotaManagerTest {
         Consumer<byte[]> consumer1 = client2.newConsumer().topic(topic1).subscriptionName(subName1).subscribe();
         Consumer<byte[]> consumer2 = client2.newConsumer().topic(topic1).subscriptionName(subName2).subscribe();
 
-        Thread producerThread = new Thread() {
-            public void run() {
-                try {
-                    barrier.await();
-                    org.apache.pulsar.client.api.Producer<byte[]> producer = createProducer(client, topic1);
-                    byte[] content = new byte[1024];
-                    for (int i = 0; i < numMsgs; i++) {
-                        producer.send(content);
-                    }
-                    producer.close();
-                } catch (Exception e) {
-                    gotException.set(true);
-                } finally {
-                    counter.countDown();
+        Thread producerThread = new Thread(() -> {
+            try {
+                barrier.await();
+                Producer<byte[]> producer = createProducer(client, topic1);
+                byte[] content = new byte[1024];
+                for (int i = 0; i < numMsgs; i++) {
+                    producer.send(content);
                 }
+                producer.close();
+            } catch (Exception e) {
+                gotException.set(true);
+            } finally {
+                counter.countDown();
             }
-        };
+        });
 
-        Thread consumerThread = new Thread() {
-            public void run() {
-                try {
-                    barrier.await();
-                    for (int i = 0; i < numMsgs; i++) {
-                        // only one consumer acknowledges the message
-                        consumer1.acknowledge(consumer1.receive());
-                        consumer2.receive();
-                    }
-                } catch (Exception e) {
-                    gotException.set(true);
-                } finally {
-                    counter.countDown();
+        Thread consumerThread = new Thread(() -> {
+            try {
+                barrier.await();
+                for (int i = 0; i < numMsgs; i++) {
+                    // only one consumer acknowledges the message
+                    consumer1.acknowledge(consumer1.receive());
+                    consumer2.receive();
                 }
+            } catch (Exception e) {
+                gotException.set(true);
+            } finally {
+                counter.countDown();
             }
-        };
+        });
 
         producerThread.start();
         consumerThread.start();
@@ -807,14 +886,14 @@ public class BacklogQuotaManagerTest {
     @Test
     public void testNoEviction() throws Exception {
         assertEquals(admin.namespaces().getBacklogQuotaMap("prop/ns-quota"),
-                Maps.newHashMap());
+                new HashMap<>());
         admin.namespaces().setBacklogQuota("prop/ns-quota",
                 BacklogQuota.builder()
                         .limitSize(10 * 1024)
                         .retentionPolicy(BacklogQuota.RetentionPolicy.consumer_backlog_eviction)
                         .build());
 
-        final String topic1 = "persistent://prop/ns-quota/topic13";
+        final String topic1 = "persistent://prop/ns-quota/topic13" + UUID.randomUUID();
         final String subName1 = "c13";
         final String subName2 = "c23";
         final int numMsgs = 10;
@@ -832,39 +911,35 @@ public class BacklogQuotaManagerTest {
         final PulsarClient client2 = PulsarClient.builder().serviceUrl(adminUrl.toString())
                 .statsInterval(0, TimeUnit.SECONDS).build();
 
-        Thread producerThread = new Thread() {
-            public void run() {
-                try {
-                    barrier.await();
-                    org.apache.pulsar.client.api.Producer<byte[]> producer = createProducer(client2, topic1);
-                    byte[] content = new byte[1024];
-                    for (int i = 0; i < numMsgs; i++) {
-                        producer.send(content);
-                    }
-                    producer.close();
-                } catch (Exception e) {
-                    gotException.set(true);
-                } finally {
-                    counter.countDown();
+        Thread producerThread = new Thread(() -> {
+            try {
+                barrier.await();
+                Producer<byte[]> producer = createProducer(client2, topic1);
+                byte[] content = new byte[1024];
+                for (int i = 0; i < numMsgs; i++) {
+                    producer.send(content);
                 }
+                producer.close();
+            } catch (Exception e) {
+                gotException.set(true);
+            } finally {
+                counter.countDown();
             }
-        };
+        });
 
-        Thread consumerThread = new Thread() {
-            public void run() {
-                try {
-                    barrier.await();
-                    for (int i = 0; i < numMsgs; i++) {
-                        consumer1.acknowledge(consumer1.receive());
-                        consumer2.acknowledge(consumer2.receive());
-                    }
-                } catch (Exception e) {
-                    gotException.set(true);
-                } finally {
-                    counter.countDown();
+        Thread consumerThread = new Thread(() -> {
+            try {
+                barrier.await();
+                for (int i = 0; i < numMsgs; i++) {
+                    consumer1.acknowledge(consumer1.receive());
+                    consumer2.acknowledge(consumer2.receive());
                 }
+            } catch (Exception e) {
+                gotException.set(true);
+            } finally {
+                counter.countDown();
             }
-        };
+        });
 
         producerThread.start();
         consumerThread.start();
@@ -875,14 +950,14 @@ public class BacklogQuotaManagerTest {
     @Test
     public void testEvictionMulti() throws Exception {
         assertEquals(admin.namespaces().getBacklogQuotaMap("prop/ns-quota"),
-                Maps.newHashMap());
+                new HashMap<>());
         admin.namespaces().setBacklogQuota("prop/ns-quota",
                 BacklogQuota.builder()
                         .limitSize(15 * 1024)
                         .retentionPolicy(BacklogQuota.RetentionPolicy.consumer_backlog_eviction)
                         .build());
 
-        final String topic1 = "persistent://prop/ns-quota/topic14";
+        final String topic1 = "persistent://prop/ns-quota/topic14" + UUID.randomUUID();
         final String subName1 = "c14";
         final String subName2 = "c24";
         final int numMsgs = 10;
@@ -903,71 +978,63 @@ public class BacklogQuotaManagerTest {
         final PulsarClient client2 = PulsarClient.builder().serviceUrl(adminUrl.toString())
                 .statsInterval(0, TimeUnit.SECONDS).build();
 
-        Thread producerThread1 = new Thread() {
-            public void run() {
-                try {
-                    barrier.await();
-                    org.apache.pulsar.client.api.Producer<byte[]> producer = createProducer(client2, topic1);
-                    byte[] content = new byte[1024];
-                    for (int i = 0; i < numMsgs; i++) {
-                        producer.send(content);
-                    }
-                    producer.close();
-                } catch (Exception e) {
-                    gotException.set(true);
-                } finally {
-                    counter.countDown();
+        Thread producerThread1 = new Thread(() -> {
+            try {
+                barrier.await();
+                Producer<byte[]> producer = createProducer(client2, topic1);
+                byte[] content = new byte[1024];
+                for (int i = 0; i < numMsgs; i++) {
+                    producer.send(content);
                 }
+                producer.close();
+            } catch (Exception e) {
+                gotException.set(true);
+            } finally {
+                counter.countDown();
             }
-        };
+        });
 
-        Thread producerThread2 = new Thread() {
-            public void run() {
-                try {
-                    barrier.await();
-                    org.apache.pulsar.client.api.Producer<byte[]> producer = createProducer(client3, topic1);
-                    byte[] content = new byte[1024];
-                    for (int i = 0; i < numMsgs; i++) {
-                        producer.send(content);
-                    }
-                    producer.close();
-                } catch (Exception e) {
-                    gotException.set(true);
-                } finally {
-                    counter.countDown();
+        Thread producerThread2 = new Thread(() -> {
+            try {
+                barrier.await();
+                Producer<byte[]> producer = createProducer(client3, topic1);
+                byte[] content = new byte[1024];
+                for (int i = 0; i < numMsgs; i++) {
+                    producer.send(content);
                 }
+                producer.close();
+            } catch (Exception e) {
+                gotException.set(true);
+            } finally {
+                counter.countDown();
             }
-        };
+        });
 
-        Thread consumerThread1 = new Thread() {
-            public void run() {
-                try {
-                    barrier.await();
-                    for (int i = 0; i < numMsgs * 2; i++) {
-                        consumer1.acknowledge(consumer1.receive());
-                    }
-                } catch (Exception e) {
-                    gotException.set(true);
-                } finally {
-                    counter.countDown();
+        Thread consumerThread1 = new Thread(() -> {
+            try {
+                barrier.await();
+                for (int i = 0; i < numMsgs * 2; i++) {
+                    consumer1.acknowledge(consumer1.receive());
                 }
+            } catch (Exception e) {
+                gotException.set(true);
+            } finally {
+                counter.countDown();
             }
-        };
+        });
 
-        Thread consumerThread2 = new Thread() {
-            public void run() {
-                try {
-                    barrier.await();
-                    for (int i = 0; i < numMsgs * 2; i++) {
-                        consumer2.acknowledge(consumer2.receive());
-                    }
-                } catch (Exception e) {
-                    gotException.set(true);
-                } finally {
-                    counter.countDown();
+        Thread consumerThread2 = new Thread(() -> {
+            try {
+                barrier.await();
+                for (int i = 0; i < numMsgs * 2; i++) {
+                    consumer2.acknowledge(consumer2.receive());
                 }
+            } catch (Exception e) {
+                gotException.set(true);
+            } finally {
+                counter.countDown();
             }
-        };
+        });
 
         producerThread1.start();
         producerThread2.start();
@@ -985,7 +1052,7 @@ public class BacklogQuotaManagerTest {
     @Test
     public void testAheadProducerOnHold() throws Exception {
         assertEquals(admin.namespaces().getBacklogQuotaMap("prop/quotahold"),
-                Maps.newHashMap());
+                new HashMap<>());
         admin.namespaces().setBacklogQuota("prop/quotahold",
                 BacklogQuota.builder()
                         .limitSize(10 * 1024)
@@ -1027,7 +1094,7 @@ public class BacklogQuotaManagerTest {
     @Test
     public void testAheadProducerOnHoldTimeout() throws Exception {
         assertEquals(admin.namespaces().getBacklogQuotaMap("prop/quotahold"),
-                Maps.newHashMap());
+                new HashMap<>());
         admin.namespaces().setBacklogQuota("prop/quotahold",
                 BacklogQuota.builder()
                         .limitSize(10 * 1024)
@@ -1065,7 +1132,7 @@ public class BacklogQuotaManagerTest {
     @Test
     public void testProducerException() throws Exception {
         assertEquals(admin.namespaces().getBacklogQuotaMap("prop/quotahold"),
-                Maps.newHashMap());
+                new HashMap<>());
         admin.namespaces().setBacklogQuota("prop/quotahold",
                 BacklogQuota.builder()
                         .limitSize(10 * 1024)
@@ -1102,10 +1169,15 @@ public class BacklogQuotaManagerTest {
         assertTrue(gotException, "backlog exceeded exception did not occur");
     }
 
-    @Test
-    public void testProducerExceptionAndThenUnblockSizeQuota() throws Exception {
+    @DataProvider(name = "dedupTestSet")
+    public static Object[][] dedupTestSet() {
+        return new Object[][] { { Boolean.TRUE }, { Boolean.FALSE } };
+    }
+
+    @Test(dataProvider = "dedupTestSet")
+    public void testProducerExceptionAndThenUnblockSizeQuota(boolean dedupTestSet) throws Exception {
         assertEquals(admin.namespaces().getBacklogQuotaMap("prop/quotahold"),
-                Maps.newHashMap());
+                new HashMap<>());
         admin.namespaces().setBacklogQuota("prop/quotahold",
                 BacklogQuota.builder()
                         .limitSize(10 * 1024)
@@ -1119,9 +1191,12 @@ public class BacklogQuotaManagerTest {
         boolean gotException = false;
 
         Consumer<byte[]> consumer = client.newConsumer().topic(topic1).subscriptionName(subName1).subscribe();
-
         byte[] content = new byte[1024];
         Producer<byte[]> producer = createProducer(client, topic1);
+
+        admin.topicPolicies().setDeduplicationStatus(topic1, dedupTestSet);
+        Thread.sleep((TIME_TO_CHECK_BACKLOG_QUOTA + 1) * 1000);
+
         for (int i = 0; i < 10; i++) {
             producer.send(content);
         }
@@ -1140,6 +1215,7 @@ public class BacklogQuotaManagerTest {
         }
 
         assertTrue(gotException, "backlog exceeded exception did not occur");
+        assertFalse(producer.isConnected());
         // now remove backlog and ensure that producer is unblocked;
 
         TopicStats stats = getTopicStats(topic1);
@@ -1156,26 +1232,46 @@ public class BacklogQuotaManagerTest {
         Exception sendException = null;
         gotException = false;
         try {
-            for (int i = 0; i < 5; i++) {
+            for (int i = 0; i < 10; i++) {
                 producer.send(content);
+                Message<?> msg = consumer.receive();
+                consumer.acknowledge(msg);
             }
         } catch (Exception e) {
             gotException = true;
             sendException = e;
         }
+        Thread.sleep((TIME_TO_CHECK_BACKLOG_QUOTA + 1) * 1000);
         assertFalse(gotException, "unable to publish due to " + sendException);
+
+        gotException = false;
+        long lastDisconnectedTimestamp = producer.getLastDisconnectedTimestamp();
+        try {
+            // try to send over backlog quota and make sure it passes
+            producer.send(content);
+            producer.send(content);
+        } catch (PulsarClientException ce) {
+            assertTrue(ce instanceof PulsarClientException.ProducerBlockedQuotaExceededException
+                    || ce instanceof PulsarClientException.TimeoutException, ce.getMessage());
+            gotException = true;
+            sendException = ce;
+        }
+        assertFalse(gotException, "unable to publish due to " + sendException);
+        assertEquals(lastDisconnectedTimestamp, producer.getLastDisconnectedTimestamp());
+
     }
 
     @Test
     public void testProducerExceptionAndThenUnblockTimeQuotaPrecise() throws Exception {
         assertEquals(admin.namespaces().getBacklogQuotaMap("prop/quotahold"),
-                Maps.newHashMap());
+                new HashMap<>());
         admin.namespaces().setBacklogQuota("prop/quotahold",
                 BacklogQuota.builder()
                         .limitTime(TIME_TO_CHECK_BACKLOG_QUOTA)
                         .retentionPolicy(BacklogQuota.RetentionPolicy.producer_exception)
                         .build(), BacklogQuota.BacklogQuotaType.message_age);
         config.setPreciseTimeBasedBacklogQuotaCheck(true);
+        @Cleanup
         final PulsarClient client = PulsarClient.builder().serviceUrl(adminUrl.toString())
                 .statsInterval(0, TimeUnit.SECONDS).build();
         final String topic1 = "persistent://prop/quotahold/exceptandunblock2";
@@ -1229,18 +1325,18 @@ public class BacklogQuotaManagerTest {
             sendException = e;
         }
         assertFalse(gotException, "unable to publish due to " + sendException);
-        client.close();
     }
 
     @Test
     public void testProducerExceptionAndThenUnblockTimeQuota() throws Exception {
         assertEquals(admin.namespaces().getBacklogQuotaMap("prop/quotahold"),
-                Maps.newHashMap());
+                new HashMap<>());
         admin.namespaces().setBacklogQuota("prop/quotahold",
                 BacklogQuota.builder()
                         .limitTime(TIME_TO_CHECK_BACKLOG_QUOTA)
                         .retentionPolicy(BacklogQuota.RetentionPolicy.producer_exception)
                         .build(), BacklogQuota.BacklogQuotaType.message_age);
+        @Cleanup
         final PulsarClient client = PulsarClient.builder().serviceUrl(adminUrl.toString())
                 .statsInterval(0, TimeUnit.SECONDS).build();
         final String topic1 = "persistent://prop/quotahold/exceptandunblock2";
@@ -1294,12 +1390,12 @@ public class BacklogQuotaManagerTest {
             sendException = e;
         }
         assertFalse(gotException, "unable to publish due to " + sendException);
-        client.close();
     }
 
     @Test(dataProvider = "backlogQuotaSizeGB", priority = 1)
     public void testBacklogQuotaInGB(boolean backlogQuotaSizeGB) throws Exception {
 
+        admin.close();
         pulsar.close();
         long backlogQuotaByte = 10 * 1024;
         if (backlogQuotaSizeGB) {
@@ -1310,12 +1406,13 @@ public class BacklogQuotaManagerTest {
         config.setBacklogQuotaDefaultRetentionPolicy(BacklogQuota.RetentionPolicy.consumer_backlog_eviction);
         pulsar = new PulsarService(config);
         pulsar.start();
+        admin = PulsarAdmin.builder().serviceHttpUrl(adminUrl.toString()).build();
 
         @Cleanup
         PulsarClient client = PulsarClient.builder().serviceUrl(pulsar.getBrokerServiceUrl()).statsInterval(0, TimeUnit.SECONDS)
                 .build();
 
-        final String topic1 = "persistent://prop/ns-quota/topic2";
+        final String topic1 = "persistent://prop/ns-quota/topic2" + UUID.randomUUID();
         final String subName1 = "c1";
         final String subName2 = "c2";
         final int numMsgs = 20;
