@@ -1,4 +1,4 @@
-/**
+/*
  * Licensed to the Apache Software Foundation (ASF) under one
  * or more contributor license agreements.  See the NOTICE file
  * distributed with this work for additional information
@@ -207,8 +207,6 @@ public class ReplicatorRateLimiterTest extends ReplicatorTestBase {
     @Test
     public void testReplicatorRatePriority() throws Exception {
         cleanup();
-        config1.setSystemTopicEnabled(true);
-        config1.setTopicLevelPoliciesEnabled(true);
         config1.setDispatchThrottlingRatePerReplicatorInMsg(100);
         config1.setDispatchThrottlingRatePerReplicatorInByte(200L);
         setup();
