@@ -1,4 +1,4 @@
-/**
+/*
  * Licensed to the Apache Software Foundation (ASF) under one
  * or more contributor license agreements.  See the NOTICE file
  * distributed with this work for additional information
@@ -47,6 +47,7 @@ import org.apache.bookkeeper.mledger.impl.PositionImpl;
 import org.apache.commons.lang3.tuple.Pair;
 import org.apache.commons.lang3.tuple.Triple;
 import org.apache.pulsar.broker.auth.MockedPulsarServiceBaseTest;
+import org.apache.pulsar.broker.service.Consumer;
 import org.apache.pulsar.broker.service.Topic;
 import org.apache.pulsar.broker.service.persistent.PersistentTopic;
 import org.apache.pulsar.client.admin.LongRunningProcessStatus;
@@ -163,6 +164,7 @@ public class CompactedTopicTest extends MockedPulsarServiceBaseTest {
 
     @Test
     public void testEntryLookup() throws Exception {
+        @Cleanup
         BookKeeper bk = pulsar.getBookKeeperClientFactory().create(
                 this.conf, null, null, Optional.empty(), null);
 
@@ -218,6 +220,7 @@ public class CompactedTopicTest extends MockedPulsarServiceBaseTest {
 
     @Test
     public void testCleanupOldCompactedTopicLedger() throws Exception {
+        @Cleanup
         BookKeeper bk = pulsar.getBookKeeperClientFactory().create(
                 this.conf, null, null, Optional.empty(), null);
 
@@ -745,6 +748,19 @@ public class CompactedTopicTest extends MockedPulsarServiceBaseTest {
                 .blockIfQueueFull(true)
                 .enableBatching(false)
                 .create();
+        // Create a consumer to generate a durable cursor, to avoid deleting the ledger before the reader receives.
+        @Cleanup
+        org.apache.pulsar.client.api.Consumer<String> consumer = pulsarClient.newConsumer(Schema.STRING)
+                .subscriptionName("sub_" + UUID.randomUUID())
+                .topic(topic)
+                .subscribe();
+        @Cleanup
+        Reader<String> reader = pulsarClient.newReader(Schema.STRING)
+                .topic(topic)
+                .startMessageIdInclusive()
+                .startMessageId(MessageId.earliest)
+                .readCompacted(true)
+                .create();
         CompletableFuture<MessageId> lastMessage = null;
         for (int i = 0; i < numMessages; ++i) {
             lastMessage = producer.newMessage().key(i + "").value(String.format("msg [%d]", i)).sendAsync();
@@ -769,14 +785,6 @@ public class CompactedTopicTest extends MockedPulsarServiceBaseTest {
         producer.flush();
         lastMessage.join();
         // For now the topic has 1000 messages in the compacted ledger and 1000 messages in the original topic.
-        @Cleanup
-        Reader<String> reader = pulsarClient.newReader(Schema.STRING)
-                .topic(topic)
-                .startMessageIdInclusive()
-                .startMessageId(MessageId.earliest)
-                .readCompacted(true)
-                .create();
-
         // Unloading the topic during reading the data to make sure the reader will not miss any messages.
         for (int i = 0; i < numMessages / 2; ++i) {
             Assert.assertEquals(reader.readNext().getValue(), String.format("msg [%d]", i));
