@@ -1,4 +1,4 @@
-/**
+/*
  * Licensed to the Apache Software Foundation (ASF) under one
  * or more contributor license agreements.  See the NOTICE file
  * distributed with this work for additional information
@@ -19,17 +19,18 @@
 package org.apache.pulsar.client.impl;
 
 import static org.mockito.ArgumentMatchers.any;
+import static org.mockito.ArgumentMatchers.anyInt;
 import static org.mockito.ArgumentMatchers.anyLong;
+import static org.mockito.ArgumentMatchers.nullable;
 import static org.mockito.Mockito.mock;
-import static org.mockito.Mockito.when;
 import static org.mockito.Mockito.verify;
+import static org.mockito.Mockito.when;
 import static org.testng.Assert.assertEquals;
 import static org.testng.Assert.assertFalse;
 import static org.testng.Assert.assertNotSame;
 import static org.testng.Assert.assertSame;
 import static org.testng.Assert.assertThrows;
 import static org.testng.Assert.assertTrue;
-
 import io.netty.buffer.ByteBuf;
 import io.netty.channel.Channel;
 import io.netty.channel.ChannelFuture;
@@ -38,7 +39,6 @@ import io.netty.channel.ChannelPromise;
 import io.netty.channel.EventLoopGroup;
 import io.netty.util.HashedWheelTimer;
 import io.netty.util.concurrent.DefaultThreadFactory;
-
 import java.lang.reflect.Field;
 import java.net.InetSocketAddress;
 import java.net.SocketAddress;
@@ -48,7 +48,6 @@ import java.util.List;
 import java.util.concurrent.CompletableFuture;
 import java.util.concurrent.ThreadFactory;
 import java.util.regex.Pattern;
-
 import lombok.Cleanup;
 import org.apache.commons.lang3.tuple.Pair;
 import org.apache.pulsar.client.api.PulsarClientException;
@@ -57,31 +56,21 @@ import org.apache.pulsar.client.impl.conf.ConsumerConfigurationData;
 import org.apache.pulsar.client.util.ExecutorProvider;
 import org.apache.pulsar.client.util.ScheduledExecutorProvider;
 import org.apache.pulsar.common.api.proto.CommandGetTopicsOfNamespace;
+import org.apache.pulsar.common.lookup.GetTopicsResult;
 import org.apache.pulsar.common.naming.NamespaceName;
 import org.apache.pulsar.common.naming.TopicName;
 import org.apache.pulsar.common.partition.PartitionedTopicMetadata;
 import org.apache.pulsar.common.util.netty.EventLoopUtil;
 import org.mockito.Mockito;
-import org.powermock.reflect.Whitebox;
 import org.testng.annotations.AfterMethod;
-import org.testng.annotations.BeforeMethod;
 import org.testng.annotations.Test;
 
 /**
  * PulsarClientImpl unit tests.
  */
 public class PulsarClientImplTest {
-    private PulsarClientImpl clientImpl;
     private EventLoopGroup eventLoopGroup;
 
-    @BeforeMethod(alwaysRun = true)
-    public void setup() throws PulsarClientException {
-        ClientConfigurationData conf = new ClientConfigurationData();
-        conf.setServiceUrl("pulsar://localhost:6650");
-        initializeEventLoopGroup(conf);
-        clientImpl = new PulsarClientImpl(conf, eventLoopGroup);
-    }
-
     private void initializeEventLoopGroup(ClientConfigurationData conf) {
         ThreadFactory threadFactory = new DefaultThreadFactory("client-test-stats", Thread.currentThread().isDaemon());
         eventLoopGroup = EventLoopUtil.newEventLoopGroup(conf.getNumIoThreads(), false, threadFactory);
@@ -89,10 +78,6 @@ public class PulsarClientImplTest {
 
     @AfterMethod(alwaysRun = true)
     public void teardown() throws Exception {
-        if (clientImpl != null) {
-            clientImpl.close();
-            clientImpl = null;
-        }
         if (eventLoopGroup != null) {
             eventLoopGroup.shutdownGracefully().get();
             eventLoopGroup = null;
@@ -101,9 +86,13 @@ public class PulsarClientImplTest {
 
     @Test
     public void testIsClosed() throws Exception {
-        assertFalse(clientImpl.isClosed());
-        clientImpl.close();
-        assertTrue(clientImpl.isClosed());
+        ClientConfigurationData conf = new ClientConfigurationData();
+        conf.setServiceUrl("pulsar://localhost:6650");
+        initializeEventLoopGroup(conf);
+        PulsarClientImpl client = new PulsarClientImpl(conf, eventLoopGroup);
+        assertFalse(client.isClosed());
+        client.close();
+        assertTrue(client.isClosed());
     }
 
     @Test
@@ -112,11 +101,14 @@ public class PulsarClientImplTest {
         LookupService lookup = mock(LookupService.class);
         when(lookup.getTopicsUnderNamespace(
                 any(NamespaceName.class),
-                any(CommandGetTopicsOfNamespace.Mode.class)))
-                .thenReturn(CompletableFuture.completedFuture(Collections.emptyList()));
+                any(CommandGetTopicsOfNamespace.Mode.class),
+                nullable(String.class),
+                nullable(String.class)))
+                .thenReturn(CompletableFuture.completedFuture(
+                        new GetTopicsResult(Collections.emptyList(), null, false, true)));
         when(lookup.getPartitionedTopicMetadata(any(TopicName.class)))
                 .thenReturn(CompletableFuture.completedFuture(new PartitionedTopicMetadata()));
-        when(lookup.getBroker(any(TopicName.class)))
+        when(lookup.getBroker(any()))
                 .thenReturn(CompletableFuture.completedFuture(
                         Pair.of(mock(InetSocketAddress.class), mock(InetSocketAddress.class))));
         ConnectionPool pool = mock(ConnectionPool.class);
@@ -131,10 +123,13 @@ public class PulsarClientImplTest {
         when(cnx.ctx()).thenReturn(ctx);
         when(cnx.sendRequestWithId(any(ByteBuf.class), anyLong()))
                 .thenReturn(CompletableFuture.completedFuture(mock(ProducerResponse.class)));
-        when(pool.getConnection(any(InetSocketAddress.class), any(InetSocketAddress.class)))
+        when(pool.getConnection(any(InetSocketAddress.class), any(InetSocketAddress.class), anyInt()))
                 .thenReturn(CompletableFuture.completedFuture(cnx));
-        Whitebox.setInternalState(clientImpl, "cnxPool", pool);
-        Whitebox.setInternalState(clientImpl, "lookup", lookup);
+
+        ClientConfigurationData conf = new ClientConfigurationData();
+        conf.setServiceUrl("pulsar://localhost:6650");
+        PulsarClientImpl client = new PulsarClientImpl(conf, eventLoopGroup, pool);
+        client.setLookup(lookup);
 
         List<ConsumerBase<byte[]>> consumers = new ArrayList<>();
         /**
@@ -143,24 +138,24 @@ public class PulsarClientImplTest {
         ConsumerConfigurationData<byte[]> consumerConf0 = new ConsumerConfigurationData<>();
         consumerConf0.setSubscriptionName("test-subscription0");
         consumerConf0.setTopicsPattern(Pattern.compile("test-topic"));
-        consumers.add((ConsumerBase) clientImpl.subscribeAsync(consumerConf0).get());
+        consumers.add((ConsumerBase<byte[]>) client.subscribeAsync(consumerConf0).get());
         /**
          * {@link org.apache.pulsar.client.impl.PulsarClientImpl#singleTopicSubscribeAsync}
          */
         ConsumerConfigurationData<byte[]> consumerConf1 = new ConsumerConfigurationData<>();
         consumerConf1.setSubscriptionName("test-subscription1");
         consumerConf1.setTopicNames(Collections.singleton("test-topic"));
-        consumers.add((ConsumerBase) clientImpl.subscribeAsync(consumerConf1).get());
+        consumers.add((ConsumerBase<byte[]>) client.subscribeAsync(consumerConf1).get());
         /**
          * {@link org.apache.pulsar.client.impl.PulsarClientImpl#multiTopicSubscribeAsync}
          */
         ConsumerConfigurationData<byte[]> consumerConf2 = new ConsumerConfigurationData<>();
         consumerConf2.setSubscriptionName("test-subscription2");
-        consumers.add((ConsumerBase) clientImpl.subscribeAsync(consumerConf2).get());
+        consumers.add((ConsumerBase<byte[]>) client.subscribeAsync(consumerConf2).get());
 
         consumers.forEach(consumer ->
                 assertNotSame(consumer.getState(), HandlerState.State.Closed));
-        clientImpl.close();
+        client.close();
         consumers.forEach(consumer ->
                 assertSame(consumer.getState(), HandlerState.State.Closed));
     }
@@ -194,27 +189,12 @@ public class PulsarClientImplTest {
         client.timer().stop();
     }
 
-    @Test(expectedExceptions = PulsarClientException.class)
-    public void testNewTransactionWhenDisable() throws Exception {
-        ClientConfigurationData conf = new ClientConfigurationData();
-        conf.setServiceUrl("pulsar://localhost:6650");
-        conf.setEnableTransaction(false);
-        PulsarClientImpl pulsarClient = null;
-        try {
-            pulsarClient = new PulsarClientImpl(conf);
-        } catch (PulsarClientException e) {
-            e.printStackTrace();
-        }
-        pulsarClient.newTransaction();
-    }
-
     @Test
-    public void testResourceCleanup() throws PulsarClientException {
-        ClientConfigurationData conf = clientImpl.conf;
+    public void testResourceCleanup() throws Exception {
+        ClientConfigurationData conf = new ClientConfigurationData();
         conf.setServiceUrl("");
         initializeEventLoopGroup(conf);
-        ConnectionPool connectionPool = new ConnectionPool(conf, eventLoopGroup);
-        try {
+        try (ConnectionPool connectionPool = new ConnectionPool(conf, eventLoopGroup)) {
             assertThrows(() -> new PulsarClientImpl(conf, eventLoopGroup, connectionPool));
         } finally {
             // Externally passed eventLoopGroup should not be shutdown.
@@ -224,7 +204,10 @@ public class PulsarClientImplTest {
 
     @Test
     public void testInitializingWithExecutorProviders() throws PulsarClientException {
-        ClientConfigurationData conf = clientImpl.conf;
+        ClientConfigurationData conf = new ClientConfigurationData();
+        conf.setServiceUrl("pulsar://localhost:6650");
+        initializeEventLoopGroup(conf);
+
         @Cleanup("shutdownNow")
         ExecutorProvider executorProvider = new ExecutorProvider(2, "shared-executor");
         @Cleanup("shutdownNow")
@@ -251,11 +234,14 @@ public class PulsarClientImplTest {
             expectedExceptionsMessageRegExp = "Both externalExecutorProvider and internalExecutorProvider must be " +
                     "specified or unspecified.")
     public void testBothExecutorProvidersMustBeSpecified() throws PulsarClientException {
-        ClientConfigurationData conf = clientImpl.conf;
+        ClientConfigurationData conf = new ClientConfigurationData();
+        conf.setServiceUrl("pulsar://localhost:6650");
+        initializeEventLoopGroup(conf);
         @Cleanup("shutdownNow")
         ExecutorProvider executorProvider = new ExecutorProvider(2, "shared-executor");
         @Cleanup
-        PulsarClientImpl client2 = PulsarClientImpl.builder().conf(conf)
+        PulsarClientImpl ignore = PulsarClientImpl.builder()
+                .conf(conf)
                 .internalExecutorProvider(executorProvider)
                 .build();
     }
